// mycompress.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "mycompress.h"


int fill_buffer(char * buffer, int buffer_length, int * total_chars_read)
{
	int chars_read;
	int eof, error;

	(*total_chars_read) = 0;

	eof = feof(input_file);
	error = ferror(input_file);
	while(buffer_length && !eof && !error)
	{
		chars_read = (int) fread(buffer, sizeof(char), buffer_length, input_file);
		
		(*total_chars_read) += chars_read;
		buffer += chars_read;
		buffer_length -= chars_read;

		eof = feof(input_file);
		error = ferror(input_file);
	}

	if (*total_chars_read)
		return 0;

	if (error)
		return -1;
	if (eof)
		return 1;

	return 0;
}

int save_buffer(char * buffer, int chars_to_write, int * total_chars_written)
{
	int chars_written;

	do
	{
		chars_written = (int) fwrite(buffer, sizeof(char), chars_to_write, output_file);

		(*total_chars_written) += chars_written;
		chars_to_write -= chars_written;
		buffer += chars_written;
	}
	while(chars_to_write && (!ferror(output_file)));
	
	if (ferror(output_file))
		return -1;

	return 0;
}

void * mymalloc(long mem_size)
{
	return malloc(mem_size);
}

void myfree(void * mem_ptr)
{
	free(mem_ptr);
}

void usage()
{
	puts("MyCompress 1.0 written by Pavel Novak, Charles University, MFF, 09/2003");
	puts("Usage: mycompress <options> <file_to_process> [<output_file>]");
	puts("");
	puts("\toptions: -c  Compress file to archive");
	puts("\t         -x  eXtract file from archive");
}

int main(int argc, char* argv[])
{
	int to_compress = 0; // 0 - compress, 1 - decompress
	char * output_filename = 0;
	
	int error = 0;

//##########################################################################
//#					region command line argument parsing				   #
//##########################################################################

for(;;) // cycle to easy go to end if error occurs
{

	if ((argc < 3) || (argc > 4) || (argv[1][0] != '-')
		|| ((argv[1][1] != '\0') && (argv[1][2] != '\0')) // more then one letter after '-'
		)
	{
		usage();
		error = 1;
		break;
	}
	
	switch (argv[1][1])
	{
	case 'c':
		to_compress = 1;
		break;
	case 'x': // to_compress set to 0 by default
		break;
	default:
		usage();
		error = 1;
		break;
	}
	if (error) break;
	
	if ((input_file = fopen(argv[2], "rb")) == 0)
	{
		printf("Unable to open input file (%s).", argv[2]);
		error = 1;
		break;
	}
	if (argc == 4)
	{
		output_filename = argv[3];
		if (!to_compress) //decompressing, skip reading input file name
		{
			char ch;
			while (fread(&ch, sizeof(char), 1, input_file) && (ch != 0));
			if (ch != 0)
			{
				printf("Unable to read input file (%s) or input file corrupted.", argv[2]);
				error = 1;
				break;
			}
		}
	}
	else // no output filename specified
	{
		if (to_compress)
		{
			if ((output_filename = (char *) malloc(strlen(argv[2]) + 5)) == 0)
			{	
				puts("Unable to allocate memory.");
				error = 1;
				break;
			}
			strcpy(output_filename, argv[2]);
			strcat(output_filename, ".myc");
		}
		else // decompressing, file name is in input file
		{
			int output_filename_length = 0;
			char buf[2];
			do
			{
				if (0 == fread(buf, sizeof(char), 1, input_file))
					break;
				output_filename_length++;
			}
			while (!feof(input_file) && !ferror(input_file) && (buf[0] != '\0'));
			
			if (buf[0] != '\0')
			{
				puts("Error input file format or error reading input file.");
				error = 1;
				break;
			}

			fseek(input_file, 0, SEEK_SET);
			if ((output_filename = (char *) malloc(output_filename_length + 1)) == 0)
			{	
				puts("Unable to allocate memory.");
				error = 1;
				break;
			}

			output_filename_length = 0;
			do
			{
				if (0 == fread(buf, sizeof(char), 1, input_file))
					break;
				output_filename[output_filename_length] = buf[0];
				output_filename_length++;
			}
			while (!feof(input_file) && !ferror(input_file) && (buf[0] != '\0'));
			if (buf[0] != '\0')
			{
				puts("Error input file format or error reading input file.");
				error = 1;
				break;
			}

		}
	}
	if ((output_file = fopen(output_filename, "wb")) == 0)
	{
		printf("Unable to open output file (%s).", output_filename);
		error = 1;
		break;
	}

//###############################################################
//#					region main program                         #
//#	now are set:											    #
//#		input_file, output_file, to_compress				    #
//###############################################################

	if (to_compress)
	{
		puts("Compressing...");
		if ((fwrite(argv[2], sizeof(char), strlen(argv[2]), output_file) != strlen(argv[2]))
			|| (fwrite("\0", sizeof(char), 1, output_file) != 1))
		{
			puts("Error writing to output file.");
			error = 1;
			break;
		}

		if (!lz77_compress(fill_buffer, save_buffer))
			puts("Compressing succesfuly completed.");
	}
	else
	{
		puts("Decompressing...");

		//TODO: extracting filename from compressed file

		if (!lz77_decompress(fill_buffer, save_buffer))
			puts("Decompressing succesfuly completed.");
	}

break;
} // for(;;)
//###############################################################
//#					region garbage collection                   #
//###############################################################

	if ((argc == 3) && (output_filename))
		free(output_filename);
	if (input_file)
		if (fclose(input_file))
			puts("Cannot close input file. This is not fatal error.");
	if (output_file)
		if (fclose(output_file))
			puts("Cannot close output file. Some data can be lost!!");

//###############################################################	
//#				    endregion garbage collection			    #
//###############################################################	

    return 0;
}
